package BasicFileSystem.KeyValue;

import Bytes.Bytes.ByteArray;
import Bytes.Bytes.ByteArrayReference;
import DataStructures.Array.Structures.Array;
import DataStructures.Array.Structures.Data;
import DataStructures.Array.Structures.DataReference;
import DataStructures.Array.Structures.Structure;
import Disk.Disk.DiskStructure;
import references.references.NumberArrayReference;
import references.references.NumberReference;
import references.references.StringArrayReference;
import references.references.StringReference;

import static Bytes.Bytes.Bytes.*;
import static DataStructures.Array.Arrays.Arrays.*;
import static DataStructures.Array.Structures.Structures.*;
import static Disk.Disk.Disk.*;
import static DiskRanges.DiskRanges.DiskRanges.*;
import static JSON.Parser.Parser.*;
import static JSON.Writer.Writer.*;
import static Unicode.Unicode.Unicode.UTF16ToUTF8;
import static Unicode.Unicode.Unicode.UTF8ToUTF16;
import static arrays.arrays.arrays.*;
import static math.math.math.*;
import static references.references.references.*;
import static strings.strings.strings.*;

public class KeyValue {
    public static boolean InitializeKeyValueStore(DiskStructure disk, double directorySize){
        Data header, root, directory;
        char[] json, blockString;
        ByteArray block;
        NumberReference entries, entrySize;
        boolean success;
        Array free;
        double diskSize;

        success = directorySize > 0d && IsInteger(directorySize);
        if(success) {
            header = CreateStructData();

            // Entries
            entries = new NumberReference();
            entrySize = new NumberReference();
            Entries(disk, entries, entrySize);
            diskSize = entries.numberValue * entrySize.numberValue;

            success = diskSize >= entrySize.numberValue + 2d*directorySize;

            if(success) {
                // First Block - Header
                AddNumberToStruct(header.structure, "DirectorySize".toCharArray(), directorySize);
                AddNumberToStruct(header.structure, "ActiveDirectory".toCharArray(), entrySize.numberValue);
                AddNumberToStruct(header.structure, "FreePosition".toCharArray(), entrySize.numberValue + 2d * directorySize);

                blockString = CreateString(entrySize.numberValue, ' ');
                json = WriteJSON(header);
                InsertStringFirst(blockString, json);
                block = UTF16ToUTF8(blockString);
                Write(disk, 0d, block);

                // Directory 1
                root = CreateStructData();
                directory = CreateStructData();
                free = CreateArray();
                AddStructToStruct(root.structure, "keys".toCharArray(), directory.structure);
                AddArrayToStruct(root.structure, "free".toCharArray(), free);

                json = WriteJSON(root);
                blockString = CreateString(directorySize, ' ');
                InsertStringFirst(blockString, json);
                block = UTF16ToUTF8(blockString);
                WriteRange(disk, entrySize.numberValue, ByteArrayLength(block), block);

                // Directory 2
                WriteRange(disk, entrySize.numberValue + ByteArrayLength(block), ByteArrayLength(block), block);

                //System.out.println((long)currentPosition/1024);
            }
        }

        return success;
    }

    public static void InsertStringFirst(char[] blockString, char[] json) {
        double i;
        for(i = 0; i < json.length; i = i + 1d){
            blockString[(int)i] = json[(int)i];
        }
    }

    public static boolean ReadHeaderAndDirectory(DiskStructure disk, NumberReference entries, NumberReference entrySize, DataReference headerRef, DataReference rootRef, StringReference message) {
        ByteArrayReference block, data;
        char [] json;
        Structure header;
        boolean success;
        double activeDirectory, directorySize;

        data = new ByteArrayReference();

        Entries(disk, entries, entrySize);

        block = new ByteArrayReference();
        block.bytes = CreateAndFillByteArray(entrySize.numberValue, 0d);

        Read(disk, 0d, block);
        json = UTF8ToUTF16(block.bytes);
        success = ReadJSON(json, headerRef, message);

        if(success){
            header = headerRef.data.structure;

            activeDirectory = GetNumberFromStruct(header, "ActiveDirectory".toCharArray());
            directorySize = GetNumberFromStruct(header, "DirectorySize".toCharArray());

            ReadRange(disk, activeDirectory, directorySize, data);
            json = UTF8ToUTF16(data.bytes);

            success = ReadJSON(json, rootRef, message);

            if(success){

            }else{
                message.string = "Error reading directory.".toCharArray();
            }
        }else{
            message.string = "Error reading header.".toCharArray();
        }

        return success;
    }

    public static boolean KeyValueAdd(DiskStructure disk, char [] key, ByteArray value, StringReference message) {
        NumberReference entries, entrySize, posRef;
        DataReference headerRef, rootRef;
        Structure header;
        Data directory, freeBlock, root;
        Array entry, free;
        boolean success, append;
        double freePosition, activeDirectory, directorySize, diskSize, newDirectory, pos;

        entries = new NumberReference();
        entrySize = new NumberReference();
        headerRef = new DataReference();
        rootRef = new DataReference();

        success = ReadHeaderAndDirectory(disk, entries, entrySize, headerRef, rootRef, message);

        if(success){
            root = rootRef.data;
            header = headerRef.data.structure;
            diskSize = entrySize.numberValue * entries.numberValue;

            freePosition = GetNumberFromStruct(header, "FreePosition".toCharArray());
            activeDirectory = GetNumberFromStruct(header, "ActiveDirectory".toCharArray());
            directorySize = GetNumberFromStruct(header, "DirectorySize".toCharArray());

            if(activeDirectory == entrySize.numberValue){
                newDirectory = activeDirectory + directorySize;
            }else{
                newDirectory = entrySize.numberValue;
            }

            directory = GetDataFromStruct(root.structure, "keys".toCharArray());
            free = GetArrayFromStruct(root.structure, "free".toCharArray());

            if(!StructHasKey(directory.structure, key)){
                // Check for free block of right size
                freeBlock = CreateArrayData();
                posRef = new NumberReference();
                append = CheckForFreeBlockOfRightSize(free, ByteArrayLength(value), freeBlock, posRef);

                entry = CreateArray();
                if(append) {
                    AddNumberToArray(entry, freePosition);
                    AddNumberToArray(entry, ByteArrayLength(value));

                    if (freePosition + ByteArrayLength(value) <= diskSize) {
                        AddArrayToStruct(directory.structure, key, entry);
                        success = WriteRange(disk, freePosition, ByteArrayLength(value), value);

                        if (success) {
                            freePosition = freePosition + ByteArrayLength(value);
                        } else {
                            message.string = "Not enough space to store value.".toCharArray();
                        }
                    } else {
                        success = false;
                        message.string = "Not enough space to store value.".toCharArray();
                    }
                }else{
                    pos = posRef.numberValue;

                    ArrayIndex(freeBlock.array, 1d).number = GetEntryLength(freeBlock.array) - ByteArrayLength(value);
                    success = WriteRange(disk, pos, ByteArrayLength(value), value);
                    if(success) {
                        AddNumberToArray(entry, pos);
                        AddNumberToArray(entry, ByteArrayLength(value));
                        AddArrayToStruct(directory.structure, key, entry);
                    }
                }
            }else{
                success = false;
                message.string = "Key is already in store.".toCharArray();
            }

            if(success){
                success = UpdateDirectoryThenHeader(disk, entrySize.numberValue, root, freePosition, directorySize, newDirectory, activeDirectory, message);
            }
        }

        return success;
    }

    public static boolean UpdateDirectoryThenHeader(DiskStructure disk, double entrySize, Data directory, double newFreePosition, double directorySize, double newDirectory, double oldDirectory, StringReference message) {
        ByteArray firstBlock, directoryData;
        char[] blockString, json, directoryString;
        Data newHeader;
        boolean success;

        json = WriteJSON(directory);
        success = json.length <= directorySize;
        if(success) {
            // Write New Directory
            directoryString = CreateString(directorySize, ' ');
            InsertStringFirst(directoryString, json);
            directoryData = UTF16ToUTF8(directoryString);
            WriteRange(disk, newDirectory, directorySize, directoryData);

            // Write Old Directory
            //WriteRange(disk, oldDirectory, directorySize, directoryData);

            // New Header
            newHeader = CreateStructData();
            AddNumberToStruct(newHeader.structure, "DirectorySize".toCharArray(), directorySize);
            AddNumberToStruct(newHeader.structure, "ActiveDirectory".toCharArray(), newDirectory);
            AddNumberToStruct(newHeader.structure, "FreePosition".toCharArray(), newFreePosition);

            blockString = CreateString(entrySize, ' ');
            json = WriteJSON(newHeader);
            InsertStringFirst(blockString, json);
            firstBlock = UTF16ToUTF8(blockString);
            Write(disk, 0d, firstBlock);
        }else{
            message.string = "Not enough space in directory.".toCharArray();
        }

        return success;
    }

    public static boolean KeyValueRemove(DiskStructure disk, char[] key, StringReference message) {
        NumberReference entrySizeRef, freePositionRef;
        DataReference headerRef, rootRef;
        boolean success;
        double directorySize, entrySize, activeDirectory, newDirectory;
        Structure header;
        Data root;

        entrySizeRef = new NumberReference();
        headerRef = new DataReference();
        rootRef = new DataReference();
        freePositionRef = new NumberReference();

        success = ReadHeaderAndDirectory(disk, new NumberReference(), entrySizeRef, headerRef, rootRef, message);

        if(success){
            entrySize = entrySizeRef.numberValue;
            header = headerRef.data.structure;
            root = rootRef.data;

            activeDirectory = GetNumberFromStruct(header, "ActiveDirectory".toCharArray());
            directorySize = GetNumberFromStruct(header, "DirectorySize".toCharArray());

            if(activeDirectory == entrySize){
                newDirectory = activeDirectory + directorySize;
            }else{
                newDirectory = entrySize;
            }

            success = KeyValueRemoveCore(key, header, root, freePositionRef, message);

            if(success){
                directorySize = GetNumberFromStruct(header, "DirectorySize".toCharArray());
                success = UpdateDirectoryThenHeader(disk, entrySize, root, freePositionRef.numberValue, directorySize, newDirectory, activeDirectory, message);
            }
        }

        return success;
    }

    public static boolean KeyValueRemoveCore(char[] key, Structure header, Data root, NumberReference freePositionRef, StringReference message) {
        Data directory;
        Array entry, free, lastFreeBlock;
        boolean success;
        double freePosition, lfpos, lflen;

        freePosition = GetNumberFromStruct(header, "FreePosition".toCharArray());

        directory = GetDataFromStruct(root.structure, "keys".toCharArray());
        free = GetArrayFromStruct(root.structure, "free".toCharArray());

        if(StructHasKey(directory.structure, key)){
            entry = GetArrayFromStruct(directory.structure, key);

            AddFreeBlock(free, entry);

            // Update free block position.
            if(ArrayLength(free) > 0d) {
                lastFreeBlock = ArrayIndex(free, ArrayLength(free) - 1d).array;
                lfpos = GetEntryPosition(lastFreeBlock);
                lflen = GetEntryLength(lastFreeBlock);
                if(lfpos + lflen == freePosition){
                    ArrayRemove(free, ArrayLength(free) - 1d);
                    freePosition = freePosition - lflen;
                }
            }

            success = RemoveFromStruct(directory.structure, key);

            freePositionRef.numberValue = freePosition;
        }else{
            success = false;
            message.string = "Key is not in store.".toCharArray();
        }

        return success;
    }

    public static double GetEntryLength(Array freeBlock) {
        return ArrayIndex(freeBlock, 1d).number;
    }

    public static void AddFreeBlock(Array free, Array block) {
        boolean higherFound;
        double i, pos, cpos, ipos, tpos, tlen, n1pos, n1len, n2pos, n2len, next;
        Array entry, tmp, n1, n2;

        pos = GetEntryPosition(block);

        // Insert in sorted list
        higherFound = false;
        ipos = 0;
        for(i = 0d; i < ArrayLength(free) && !higherFound; i = i + 1d){
            entry = ArrayIndex(free, i).array;
            cpos = GetEntryPosition(entry);

            if(pos < cpos){
                higherFound = true;
                ipos = i;
            }
        }

        AddArrayToArray(free, block);

        tmp = block;

        if(higherFound){
            for(i = ipos; i < ArrayLength(free); i = i + 1d){
                entry = ArrayIndex(free, i).array;
                tpos = GetEntryPosition(entry);
                tlen = GetEntryLength(entry);
                ArrayIndex(entry, 0d).number = GetEntryPosition(tmp);
                ArrayIndex(entry, 1d).number = GetEntryLength(tmp);
                ArrayIndex(tmp, 0d).number = tpos;
                ArrayIndex(tmp, 1d).number = tlen;
            }
        }

        // Combine adjacent blocks
        next = 1d;
        for(i = 0; i < ArrayLength(free) - next; i = i + 1d){
            n1 = ArrayIndex(free, i).array;
            n2 = ArrayIndex(free, i+next).array;
            n1pos = GetEntryPosition(n1);
            n1len = GetEntryLength(n1);
            n2pos = GetEntryPosition(n2);
            n2len = GetEntryLength(n2);

            if(n1pos + n1len == n2pos){
                ArrayIndex(n1, 1d).number = n1len + n2len;
                ArrayIndex(n2, 1d).number = 0;
                i = i - 1d;
                next = next + 1d;
            }
        }

        // Remove entries of zero size
        for(i = 0d; i < ArrayLength(free); i = i + 1d){
            n1 = ArrayIndex(free, i).array;
            n1len = GetEntryLength(n1);

            if(n1len == 0d){
                ArrayRemove(free, i);
                i = i - 1d;
            }
        }
    }

    public static double GetEntryPosition(Array entry) {
        return ArrayIndex(entry, 0d).number;
    }

    public static boolean RemoveFromStruct(Structure structure, char[] key) {
        double i, p;
        char [] ckey;
        boolean found;

        found = false;
        p = 0d;

        for(i = 0d; i < ArrayLength(structure.keys) && !found; i = i + 1d){
            ckey = ArrayIndex(structure.keys, i).string;
            if(StringsEqual(ckey, key)){
                p = i;
                found = true;
            }
        }

        if(found){
            ArrayRemove(structure.keys, p);
            ArrayRemove(structure.values, p);
        }

        return found;
    }

    public static boolean KeyValueRead(DiskStructure disk, char[] key, ByteArrayReference data, StringReference message) {
        DataReference rootRef;
        Data root, directory;
        boolean success;
        double pos, len;
        Array entry;

        rootRef = new DataReference();
        success = ReadHeaderAndDirectory(disk, new NumberReference(), new NumberReference(), new DataReference(), rootRef, message);

        if(success){
            root = rootRef.data;

            directory = GetDataFromStruct(root.structure, "keys".toCharArray());

            if(StructHasKey(directory.structure, key)) {
                entry = GetArrayFromStruct(directory.structure, key);
                pos = GetEntryPosition(entry);
                len = GetEntryLength(entry);

                success = ReadRange(disk, pos, len, data);
            } else {
                success = false;
                message.string = "Key is not in store.".toCharArray();
            }
        }

        return success;
    }

    public static boolean KeyValueList(DiskStructure disk, StringArrayReference dir, StringReference message) {
        DataReference rootRef;
        Data  directory;
        Structure root;
        boolean success;

        rootRef = new DataReference();

        success = ReadHeaderAndDirectory(disk, new NumberReference(), new NumberReference(), new DataReference(), rootRef, message);

        if(success){
            root = rootRef.data.structure;
            directory = GetDataFromStruct(root, "keys".toCharArray());
            dir.stringArray = GetStructKeys(directory.structure);
        }

        return success;
    }

    public static boolean KeyValueListStartsWith(DiskStructure disk, char[] prefix, StringArrayReference dir, StringReference message) {
        boolean success;
        StringArrayReference dir2;
        double i;
        char [] key;
        Array filtered;

        filtered = CreateArray();
        dir2 = new StringArrayReference();
        success = KeyValueList(disk, dir2, message);

        if(success){
            for(i = 0d; i < dir2.stringArray.length; i = i + 1d){
                key = dir2.stringArray[(int)i].string;
                if(StartsWith(key, prefix)){
                    AddStringToArray(filtered, key);
                }
            }

            dir.stringArray = new StringReference[(int)ArrayLength(filtered)];
            for(i = 0d; i < ArrayLength(filtered); i = i + 1d){
                dir.stringArray[(int)i] = new StringReference();
                dir.stringArray[(int)i].string = ArrayIndex(filtered, i).string;
            }
        }

        return success;
    }

    public static boolean KeyValueUpdate(DiskStructure disk, char [] key, ByteArray value, StringReference message) {
        NumberReference entries, entrySizeRef, posRef, freePositionRef;
        DataReference headerRef, rootRef;
        Structure header;
        Data directory, freeBlock, root;
        Array entry, free;
        boolean success, append;
        double freePosition, activeDirectory, directorySize, diskSize, newDirectory, pos, entrySize;

        entries = new NumberReference();
        entrySizeRef = new NumberReference();
        headerRef = new DataReference();
        rootRef = new DataReference();

        success = ReadHeaderAndDirectory(disk, entries, entrySizeRef, headerRef, rootRef, message);

        if(success){
            entrySize = entrySizeRef.numberValue;
            root = rootRef.data;
            header = headerRef.data.structure;
            diskSize = entrySizeRef.numberValue * entries.numberValue;

            freePosition = GetNumberFromStruct(header, "FreePosition".toCharArray());
            activeDirectory = GetNumberFromStruct(header, "ActiveDirectory".toCharArray());
            directorySize = GetNumberFromStruct(header, "DirectorySize".toCharArray());

            if(activeDirectory == entrySizeRef.numberValue){
                newDirectory = activeDirectory + directorySize;
            }else{
                newDirectory = entrySizeRef.numberValue;
            }

            directory = GetDataFromStruct(root.structure, "keys".toCharArray());
            free = GetArrayFromStruct(root.structure, "free".toCharArray());

            // Check for free block of right size
            freeBlock = CreateArrayData();
            posRef = new NumberReference();
            append = CheckForFreeBlockOfRightSize(free, ByteArrayLength(value), freeBlock, posRef);

            entry = CreateArray();
            if(append) {
                AddNumberToArray(entry, freePosition);
                AddNumberToArray(entry, ByteArrayLength(value));

                if (freePosition + ByteArrayLength(value) <= diskSize) {

                    success = WriteRange(disk, freePosition, ByteArrayLength(value), value);

                    if (success) {
                        freePosition = freePosition + ByteArrayLength(value);
                    } else {
                        message.string = "Not enough space to store value.".toCharArray();
                    }
                } else {
                    success = false;
                    message.string = "Not enough space to store value.".toCharArray();
                }
            }else{
                pos = posRef.numberValue;

                ArrayIndex(freeBlock.array, 1d).number = GetEntryLength(freeBlock.array) - ByteArrayLength(value);
                success = WriteRange(disk, pos, ByteArrayLength(value), value);
                if(success) {
                    AddNumberToArray(entry, pos);
                    AddNumberToArray(entry, ByteArrayLength(value));
                }
            }

            if(success){
                freePositionRef = CreateNumberReference(freePosition);
                success = KeyValueRemoveCore(key, header, root, freePositionRef, message);

                if(success){
                    AddArrayToStruct(directory.structure, key, entry);
                    freePosition = freePositionRef.numberValue;
                    directorySize = GetNumberFromStruct(header, "DirectorySize".toCharArray());
                    success = UpdateDirectoryThenHeader(disk, entrySize, root, freePosition, directorySize, newDirectory, activeDirectory, message);
                }
            }
        }

        return success;
    }

    public static boolean CheckForFreeBlockOfRightSize(Array free, double size, Data freeBlock, NumberReference posRef){
        boolean append;
        double i, len;

        append = true;
        posRef.numberValue = 0d;
        for(i = 0; i < ArrayLength(free) && append; i = i + 1d){
            freeBlock.array = ArrayIndex(free, i).array;
            posRef.numberValue = GetEntryPosition(freeBlock.array);
            len = GetEntryLength(freeBlock.array);
            if(size <= len){
                append = false;
            }
        }

        return append;
    }

    public static boolean KeyValueFree(DiskStructure disk, NumberReference diskSpace, NumberReference directorySpace, StringReference message) {
        char [] json;
        NumberReference entries, entrySize;
        DataReference headerRef, rootRef;
        Structure header;
        boolean success;
        double directorySize, diskSize, freePosition, i;
        Array free, freeBlock;

        entries = new NumberReference();
        entrySize = new NumberReference();
        headerRef = new DataReference();
        rootRef = new DataReference();

        success = ReadHeaderAndDirectory(disk, entries, entrySize, headerRef, rootRef, message);

        if(success){
            header = headerRef.data.structure;
            diskSize = entries.numberValue * entrySize.numberValue;
            json = WriteJSON(rootRef.data);

            freePosition = GetNumberFromStruct(header, "FreePosition".toCharArray());
            directorySize = GetNumberFromStruct(header, "DirectorySize".toCharArray());

            directorySpace.numberValue = directorySize - json.length;
            diskSpace.numberValue = diskSize - freePosition;

            free = GetArrayFromStruct(rootRef.data.structure, "free".toCharArray());

            for(i = 0d; i < ArrayLength(free); i = i + 1d){
                freeBlock = ArrayIndex(free, i).array;
                diskSpace.numberValue = diskSpace.numberValue + GetEntryLength(freeBlock);
            }
        }

        return success;
    }
}
